#include "str_da.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// `strdup`s each element on insertion
#define CLONE_ON_INSERT

// `free`s each element when it is removed or overwritten
#define FREE_ON_REMOVE

// leaves the data pointer NULL rather than heap allocating 0 bytes
#define ZERO_CAP_OPTIMIZATION

typedef char* element_t;

struct str_da
{
    size_t capacity;
    size_t size;
    element_t* data;
};

str_da_t str_da_create(size_t capacity)
{
    str_da_t result = malloc(sizeof(struct str_da));
    if (result == NULL) goto fail1;

    result->capacity = capacity;
    result->size = 0;

#ifdef ZERO_CAP_OPTIMIZATION
    if (capacity == 0) {
        result->data = NULL;
        return result;
    }
#endif

    result->data = malloc(capacity * sizeof(element_t));
    if (result->data == NULL) goto fail2;

    return result;

    fail2:
    free(result);

    fail1:
    return NULL;
}

void str_da_destroy(str_da_t da)
{
#ifdef FREE_ON_REMOVE
    for (size_t i = 0; i < da->size; ++i)
        free(da->data[i]);
#endif

    free(da->data);
    free(da);
}

size_t str_da_size(str_da_t da)
{
    return da->size;
}

element_t str_da_get(str_da_t da, size_t ix)
{
    return da->data[ix];
}

static char* clone(char const* who, char *s)
{
    if (s != NULL) {
        s = strdup(s);
        if (s == NULL) {
            perror(who);
            exit(1);
        }
    }

    return s;
}

void str_da_set(str_da_t da, size_t ix, element_t val)
{
#ifdef FREE_ON_REMOVE
    free(da->data[ix]);
#endif
#ifdef CLONE_ON_INSERT
    val = clone("str_da_set", val);
#endif
    da->data[ix] = val;
}

static size_t max(size_t a, size_t b)
{
    return a < b ? b : a;
}

bool str_da_push_back(str_da_t da, element_t val)
{
    if (da->size == da->capacity) {
        size_t new_cap = max(da->size + 1, 2 * da->capacity);
        element_t* new_data = realloc(da->data, new_cap * sizeof(element_t));
        if (new_data == NULL) return false;

        da->capacity = new_cap;
        da->data = new_data;
    }

#ifdef CLONE_ON_INSERT
    val = clone("str_da_push_back", val);
#endif

    da->data[da->size++] = val;
    return true;
}

void str_da_pop_back(str_da_t da)
{
    --da->size;
#ifdef FREE_ON_REMOVE
    free(da->data[da->size]);
#endif
}

