#pragma once

#include <stddef.h>
#include <stdbool.h>

// Dynamic arrays of strings are represented as a pointer to an
// incomplete implementation type.
typedef struct str_da* str_da_t;

// Allocates a new array with the given initial capacity.
str_da_t str_da_create(size_t capacity);

// Deallocates the given array, recovering its storage.
void str_da_destroy(str_da_t);

// Returns the number of elements of the given array.
size_t str_da_size(str_da_t);

// Gets the `i`th element of the array.
char* str_da_get(str_da_t, size_t);

// Sets the `i`th element of the array.
void str_da_set(str_da_t, size_t, char*);

// Increments the size of the array and places the given string
// in the new position.
// Returns true for success, or false if the array needs to grow
// but can't.
bool str_da_push_back(str_da_t, char*);

// Removes the last element of the array, decrementing its size.
void str_da_pop_back(str_da_t);
